@@ -55,7 +55,7 @@ XrdCl::Access::Mode file_mode_to_xrdcl_access( mode_t mode )
   if( mode & S_IWOTH )
     xrdcl_mode |= XrdCl::Access::OW;
   if( mode & S_IXOTH )
-    xrdcl_mode |= XrdCl::Access::OX;   
+    xrdcl_mode |= XrdCl::Access::OX;
 
   return xrdcl_mode;
 }
@@ -92,7 +92,7 @@ static std::string query_args(gfal2_context_t context, const char *url)
     std::ostringstream args;
 
     if (ucert) {
-    
+
         if (!ukey) {
             ukey = ucert;
         }
@@ -107,9 +107,9 @@ static std::string query_args(gfal2_context_t context, const char *url)
            args << "xrd.gsiusrcrt=" << ucert << '&' << "xrd.gsiusrkey=" << ukey;
            prev_args = true;
         }
-     
+
         g_free(ucert);
-    
+
         if (ucert != ukey)
             g_free(ukey);
     }
@@ -262,7 +262,7 @@ void collapse_slashes(std::string& path)
 
 // Mask network error codes as ECOMM
 #ifndef ECOMM
-#define ECOMM 70
+#define ECOMM EIO
 #endif
 
 static int mask_network_errno(int errc)
@@ -284,8 +284,12 @@ static int mask_network_errno(int errc)
 
 
 // Copied from xrootd/src/XrdPosix/XrdPosixMap.cc
+#ifndef ENOSR
+#define ENOSR ENOSPC
+#endif
+
 #ifndef ECHRNG
-#define ECHRNG 44
+#define ECHRNG EINVAL
 #endif
 
 static int map_status_code_to_errno(int code)
